/* IM-JA Japanese Input Method Module for GTK-2.0
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Botond Botyanszki <boti@rocketmail.com>
 *
 * Based on code by:
 *          Owen Taylor <otaylor@redhat.com>
 *          Yusuke Tabata <tee@kuis.kyoto-u.ac.jp>
 *          Yukihiro Nakai <ynakai@redhat.com>
 *          
 */

#include <string.h>
#include <stdio.h>
#include <canna/jrkanji.h>
#include <canna/keydef.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include "im-ja.h"
#include "common.h"
#include "conf.h"
#include "error.h"

/* * * * * * * * * * * */

void im_ja_show_candwin(IMContextJA *cn, int selected);
void im_ja_hide_candwin(IMContextJA *cn);
void im_ja_set_preedit_string(IMContextJA *context, gchar *str, gint reverse_start, gint reverse_end);
void im_ja_reset(IMContextJA *context);
void im_ja_free_candidate_list(IMContextJA *context);

void roma2kana_canna(IMContextJA *context, guchar newinput);
gchar *euc2utf8 (const gchar *str);
gchar *utf82euc (const gchar *str);
gboolean im_ja_canna_filter_keypress(IMContextJA *context, GdkEventKey *key);
void im_ja_canna_reset(IMContextJA *cn);
void im_ja_canna_commit(IMContextJA *cn);

extern IMJAConfig cfg;

/* * * * * * * * * * * */

void im_ja_canna_force_init(IMContextJA *cn) {
  cn->ksv.ks = &(cn->ks);
  cn->ksv.buffer = g_new0(gchar, BUFFERSIZE);
  cn->ksv.bytes_buffer = BUFFERSIZE;
  cn->ksv.val = CANNA_MODE_HenkanMode;
  jrKanjiControl(0, KC_CHANGEMODE, (void*)&cn->ksv);
  jrKanjiControl(0, KC_SETWIDTH, (gchar*) 100);
  //  printf("canna init.\n");
}

void im_ja_canna_reset(IMContextJA *cn) {
  jrKanjiControl(0, KC_KILL, (void*)&cn->ksv);
}

void im_ja_canna_commit(IMContextJA *cn) {
  jrKanjiControl(0, KC_KAKUTEI, (void*)&cn->ksv);
}

gboolean im_ja_canna_filter_keypress(IMContextJA *cn, GdkEventKey *key) {
  if( key->type == GDK_KEY_RELEASE ) {
    return FALSE;
  };

  /* ENTER */
  if (key->keyval == GDK_Return ) {
    if (strlen(cn->preedit_buf) > 0) { //Commit preedit buffer
      g_signal_emit_by_name(cn, "commit", cn->preedit_buf);
      im_ja_reset(cn);
      im_ja_canna_commit(cn);
      return TRUE;
    }
    return FALSE; // Insert the Enter
  }
  /* BACKSPACE */
  
  if (key->keyval == GDK_BackSpace ) {
    if (strlen(cn->preedit_buf) > 0) { //Delete preedit
      roma2kana_canna(cn, 0x08);
      g_signal_emit_by_name(cn, "preedit_changed");
      return TRUE;
    }
    return FALSE; //Delete normal
  }
  
  /* SPACE */
  if (key->keyval == GDK_space) { 
    if (strlen(cn->preedit_buf) == 0) {  //Insert a space otherwise do the conversion
      g_signal_emit_by_name(cn, "commit", " ");
    }
    else {
      cn->cand_stat++;
      roma2kana_canna(cn, ' ');
      g_signal_emit_by_name(cn, "preedit_changed");
    }
    return TRUE;
  }

  /* Convert */
  if (ishotkey(key, CONVERT, &cfg) == TRUE) {
    if (strlen(cn->preedit_buf) == 0) return FALSE;
    cn->cand_stat++;
    roma2kana_canna(cn, ' ');
    g_signal_emit_by_name(cn, "preedit_changed");
    return TRUE;
  }

  if (ishotkey(key, UNCONVERT, &cfg) == TRUE) {
    if (strlen(cn->preedit_buf) == 0) return FALSE;
    IM_JA_DEBUG("Canna: unconvert\n");
    roma2kana_canna(cn, 0x07 );
    g_signal_emit_by_name(cn, "preedit_changed");
    return TRUE;
  }  

  /* Expand current bunsetsu */
  if (ishotkey(key, EXPAND_BUNSETSU, &cfg) == TRUE) {
    if (strlen(cn->preedit_buf) == 0) return FALSE;
    roma2kana_canna(cn, 0x0f /*CANNA_KEY_Shift_Right*/ ); //0x0f);
    g_signal_emit_by_name(cn, "preedit_changed");
    IM_JA_DEBUG("Expand current bunsetsu\n");
    return TRUE;
  } 
  
  /* Shrink current bunsetsu */
  if (ishotkey(key, SHRINK_BUNSETSU, &cfg) == TRUE) {
    if (strlen(cn->preedit_buf) == 0) return FALSE;
    roma2kana_canna(cn, 0x09 /*CANNA_KEY_Shift_Left*/ ); //0x09);
    g_signal_emit_by_name(cn, "preedit_changed");
    IM_JA_DEBUG("Shrink current bunsetsu\n");
    return TRUE;
  }
  
  /* Select next bunsetsu */
  if (ishotkey(key, NEXT_BUNSETSU, &cfg) == TRUE) {
    if (strlen(cn->preedit_buf) == 0) return FALSE;
    roma2kana_canna(cn, 0x06 /*CANNA_KEY_Right*/);
    g_signal_emit_by_name(cn, "preedit_changed");
    IM_JA_DEBUG("Select next bunsetsu\n");
    return TRUE;
  }
  
  /* Select prev bunsetsu */
  if (ishotkey(key, PREV_BUNSETSU, &cfg) == TRUE) {
     if (strlen(cn->preedit_buf) == 0) return FALSE;
    roma2kana_canna(cn, 0x02 /*CANNA_KEY_Left*/);
    g_signal_emit_by_name(cn, "preedit_changed");
    IM_JA_DEBUG("Select prev bunsetsu\n");
    return TRUE;
  }

  /* Select previous candidate */
  if (ishotkey(key, PREV_CANDIDATE, &cfg) == TRUE) {
    if (strlen(cn->preedit_buf) == 0) return FALSE;
    roma2kana_canna(cn, 0x10 /*CANNA_KEY_Up*/);
    g_signal_emit_by_name(cn, "preedit_changed");
    IM_JA_DEBUG("Select previous candidate\n");
    return TRUE;
  }

   /* Select next candidate */
  if (ishotkey(key, NEXT_CANDIDATE, &cfg) == TRUE) {
    if (strlen(cn->preedit_buf) == 0) return FALSE;
    roma2kana_canna(cn, 0x0e /*CANNA_KEY_Down*/);
    g_signal_emit_by_name(cn, "preedit_changed");
    IM_JA_DEBUG("Select next candidate\n");
    return TRUE;
  }

  /* REGULAR CHAR KEYS */
  if (key->keyval >= GDK_exclam && key->keyval <= GDK_overline) { //else if (key->keyval >= GDK_a && key->keyval <= GDK_z) { 
    if (cn->cand_stat != 0) { //if we have a candidate, then commit it.
      //  printf("candidate.\n");
      im_ja_canna_reset(cn);
      g_signal_emit_by_name(cn, "commit", cn->preedit_buf);
      im_ja_reset(cn);
    }
    roma2kana_canna(cn, key->string[0]); //gdk_keyval_name(key->keyval)[0]); 
    g_signal_emit_by_name(cn, "preedit_changed");
    return TRUE;
  }

  return FALSE; //We didn't handle the keypress
}

void roma2kana_canna(IMContextJA *cn, guchar newinput) {
  gint nbytes;
  gchar *tmpptr;

  IM_JA_DEBUG("roma2kana newchar: 0x%X\n", newinput);
  im_ja_free_candidate_list(cn);
  
  nbytes = jrKanjiString(0, (int)newinput, cn->kakutei_buf, BUFFERSIZE, &cn->ks); //Do the conversion

  if (cn->ks.length > 0) { //conversion succeeded 
    IM_JA_DEBUG("Canna: conversion OK\n");
    memset(cn->preedit_buf, 0, BUFFERSIZE);
    tmpptr = g_new0(gchar, BUFFERSIZE);
    strncpy(tmpptr, cn->ks.echoStr, cn->ks.length);
    im_ja_set_preedit_string(cn, euc2utf8(tmpptr), 0, 0);
    g_free(tmpptr);
    tmpptr = g_strndup(cn->ks.echoStr, cn->ks.revPos);
    cn->preedit_reverse_start = strlen(euc2utf8(tmpptr));
    g_free(tmpptr);
    tmpptr = g_strndup(cn->ks.echoStr, cn->ks.revPos + cn->ks.revLen);
    cn->preedit_reverse_end = strlen(euc2utf8(tmpptr));
    g_free(tmpptr);
  } 
  else {
    if (cn->ks.length == 0) { //conversion failed
	  IM_JA_DEBUG("Canna: conversion Failed (error=%s,nbytes=%d)!\n",jrKanjiError,nbytes);
      im_ja_set_preedit_string(cn, NULL, 0, 0);
      memset(cn->kakutei_buf, 0, BUFFERSIZE);
    }
  }
  if (cn->ks.info & KanjiGLineInfo) { //If we get candidates
    gchar *tmpcand = NULL;
    //printf("GLineMode: %s\n", cn->ks.mode);
    //printf("candidate:%s pos: %d, length: %d\n", cn->ks.echoStr, cn->ks.revPos, cn->ks.revLen);
    //cn->cand_stat++;
    tmpptr = euc2utf8(g_strdup(cn->ks.gline.line));
    IM_JA_DEBUG("Candidates: (%s)\n", tmpptr );
    /* Canna returns one string with all the candidates.
     * The candidates are seperated with a zenkaku space (hex: e3 80 80)
     * and the last candidate is followed by spaces, followed by candidate count
     *
     * The following loop extracts all the candidates except the last one.
     */
    while (TRUE) {
      //skip the candidate number
      tmpcand = g_utf8_next_char(tmpptr);
      IM_JA_DEBUG("tmpcand = %s\n", tmpcand );
      tmpptr = g_strstr_len(tmpptr, strlen(tmpptr), "\xe3\x80\x80"); //find zenkaku space
      if (tmpptr != NULL) {
	tmpptr[0] = 0;
	tmpptr++;
	tmpptr = g_utf8_find_next_char(tmpptr, NULL);
      }
      else break;
     
      /* Dupe the string, because im_ja_free_candidate_list will try to free each
       * candidate. */
      cn->candidate_list = g_list_append(cn->candidate_list, g_strdup(tmpcand));
    }
    /* remove space from the end */
    
    for (tmpptr = tmpcand; *tmpptr != 0; tmpptr++) {
      if (*tmpptr == ' ') { 
	*tmpptr = 0; //FIXME: we need to get the position: 2/4, 
      }
    }
    /* Add the final candidate */
    cn->candidate_list = g_list_append(cn->candidate_list, g_strdup(tmpcand));

    im_ja_show_candwin(cn, 0);
  }

  //  printf("preedit_buf(%d): >%s<\n", strlen(cn->preedit_buf), utf82euc(cn->preedit_buf));
  //printf("kakutei: >%s<\n", cn->kakutei_buf);
  if (strlen(cn->kakutei_buf) == 1 && cn->kakutei_buf[0] == newinput) {
    cn->kakutei_buf[0] = '\0';
  }
  //printf("roma2kana done.\n");
}

