/* -*- Mode: C; tab-width: 2; indent-tabs-mode: t; c-basic-offset: 2 -*- */
#ifndef _IM_JA_H_
#define _IM_JA_H_

#include <gtk/gtk.h>
#include <gtk/gtkimmodule.h>

#define BUFFERSIZE 1024

extern GType type_im_context_im_ja;

#define TYPE_IM_CONTEXT_IM_JA              type_im_context_im_ja
#define IM_CONTEXT_IM_JA(obj)              (GTK_CHECK_CAST ((obj), TYPE_IM_CONTEXT_IM_JA, IMContextJA))
#define IM_CONTEXT_IM_JA_CLASS(klass)      (GTK_CHECK_CLASS_CAST ((klass), TYPE_IM_CONTEXT_IM_JA, IMContextJAClass))
#define IS_IM_CONTEXT_IM_JA(obj)           (GTK_CHECK_TYPE ((obj), TYPE_IM_CONTEXT_IM_JA))
#define IS_IM_CONTEXT_IM_JA_CLASS(klass)   (GTK_CHECK_CLASS_TYPE ((klass), TYPE_IM_CONTEXT_IM_JA))
#define IM_CONTEXT_IM_JA_GET_CLASS(obj)    (GTK_CHECK_GET_CLASS ((obj), TYPE_IM_CONTEXT_IM_JA, IMContextJAClass))

//Status widget types
enum {
  STATUS_BELOW_CURSOR,
  STATUS_WINDOW_BOTTOM_LEFT,
  STATUS_FIXED_COORD,
  STATUS_DISABLED
};

//Candwin styles
enum {
	CANDWIN_STYLE_MENU,
	CANDWIN_STYLE_TABLE
};

// Conversion engines
enum { 
  CONV_ENGINE_CANNA,
  CONV_ENGINE_WNN,
	CONV_ENGINE_KANJIPAD,
  CONV_ENGINE_TOTAL
  /* TO BE IMPLEMENTED: 
  IM_JA_ANTHY_INPUT ,
  IM_JA_SKK_INPUT,
  IM_JA_ATOK_INPUT,
  IM_JA_XIM_INPUT,
	*/
};

enum {
  IM_JA_RAW_INPUT,
  IM_JA_HIRAGANA_INPUT,
  IM_JA_KATAKANA_INPUT,
  IM_JA_HALFKATA_INPUT,
  IM_JA_ZENKAKU_INPUT,
  IM_JA_KANJIPAD_INPUT,
	
  IM_JA_INPUT_METHODS_TOTAL

  /* TO BE IMPLEMENTED: 
  IM_JA_SYMBOL_INPUT,
  IM_JA_HEX_INPUT,
  IM_JA_TOROKU_INPUT,
  IM_JA_BUSHU_INPUT
  */
};

// Keys
enum {
  START_CONFIGURATOR,
  NEXT_INPUT_MODE,
  PREV_INPUT_MODE,
  RAW_INPUT_MODE,
  HIRAGANA_INPUT_MODE,
  KATAKANA_INPUT_MODE,
  HALFKATA_INPUT_MODE,
  ZENKAKU_INPUT_MODE,
  CANNA_INPUT_MODE,
  WNN_INPUT_MODE,
  KANJIPAD_INPUT_MODE,
  UNCONVERT,
  CONVERT,
  PREV_CANDIDATE,
  NEXT_CANDIDATE,
  LIST_CANDIDATES,
  PREV_BUNSETSU,
  NEXT_BUNSETSU,
  EXPAND_BUNSETSU,
  SHRINK_BUNSETSU,
  TOGGLE_PREEDIT_WINDOW,
	COMMIT_PREEDIT,
	SYMBOL_INPUT,
	UNICODE_INPUT,
	JIS_CODE_INPUT,
  HOTKEYS_TOTAL
};

typedef struct _IMContextJA IMContextJA;
typedef struct _IMContextJAClass IMContextJAClass;

#include "statuswin.h"
#include "preeditwin.h"
#include "candwin.h"

#ifndef DISABLE_WNN
#include "wnn.h"
#endif

struct _IMContextJAClass {
  GtkIMContextClass parent_class;
};

struct _IMContextJA {
  GtkIMContext parent_object;

	//guint client_win_destroy_handler;
	gboolean finalized;
  GdkWindow *client_gdk;
  GtkWidget *client_gtk;
  GdkWindow *toplevel_gdk;
  GtkWidget *toplevel_gtk;

  GtkWidget *kanjipad;
	StatusWin *status_win;
	PreeditWin *preedit_win;
	CandidateWin *candidate_win;

  gint candwin_pos_offset_x;
  gint candwin_pos_offset_y;
  gint preedit_win_pos_offset_x;
  gint preedit_win_pos_offset_y;
	gint cursor_pos_offset_x; 
	gint cursor_pos_offset_y;
  gboolean update_candwin_pos;
  gboolean update_preedit_pos;

	GdkColor original_colors[2]; /* original client window widget style: fg and bg */

  GList *candidate_list;
  gboolean show_first;

  gboolean preedit_win_on;

  /* Preedit string */
  gchar *preedit_buf;
  gchar *preedit_insert;
  gint preedit_reverse_start;
  gint preedit_reverse_end;

  gint cursor_ndx;
  gint cursor_wnn;

  gint cand_stat;
  gboolean show_candidates;
#ifndef DISABLE_WNN
  wnnStat wnn_stat;
#endif
  gint input_method;
	gint conv_engine;
	gboolean conv_engine_initialized;
  void *canna_rk_ctx;
};

/* Function prototypes */
GtkIMContext *im_ja_new(void);
void im_ja_register_type(GTypeModule *module);

void im_ja_next_input_method(IMContextJA *context);
void im_ja_prev_input_method(IMContextJA *context);
gboolean im_ja_set_input_method(IMContextJA *context, int input_method);
void im_ja_reset(IMContextJA *context);
void im_ja_set_preedit_string(IMContextJA *context, gchar *str, gint reverse_start, gint reverse_end);
void im_ja_free_candidate_list(IMContextJA *context);
void im_ja_commit(IMContextJA *cn);
void im_ja_grab_add(GtkWidget *widget, GdkEvent *event, IMContextJA* cn);
void im_ja_grab_remove(GtkWidget *widget, GdkEvent *event, IMContextJA* cn);
void im_ja_input_utf8(gchar *strg, IMContextJA *context);
#endif
