/* -*- Mode: C; tab-width: 2; indent-tabs-mode: t; c-basic-offset: 2 -*- */
/* IM-JA Japanese Input Method Module for GTK-2.0
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Botond Botyanszki <boti@rocketmail.com>
 *
 */

#include <config.h>
#include <gtk/gtkimmodule.h>
#include <gtk/gtk.h>
#include <gmodule.h>

#include <string.h>

#include "im-ja.h"
#include "error.h"
#include "nls.h"

extern GType gtk_im_module_get_type(void);
#define GTK_TYPE_IM_MODULE (gtk_im_module_get_type())
#define GTK_IM_MODULE(im_module) (G_TYPE_CHECK_INSTANCE_CAST((im_module), GTK_TYPE_IM_MODULE, GtkIMModule))

typedef struct _GtkIMModule {
	GTypeModule parent_instance;
	GModule *library;
} GtkIMModule;

static const GtkIMContextInfo im_ja_info = { 
	"im-ja",           /* ID */
	N_("Japanese"),    /* Human readable name */
	GETTEXT_PACKAGE,   /* Translation domain */
	IM_JA_LOCALE_DIR,  /* Dir for bindtextdomain (not strictly needed for "gtk+") */
	"ja"               /* Languages for which this module is the default */
};

static const GtkIMContextInfo *info_list[] = {
	&im_ja_info
};

void im_module_init(GTypeModule *module) {
	IM_JA_DEBUG("im_module_init()\n");
	GtkIMModule *im_module = GTK_IM_MODULE(module);
	g_module_make_resident(im_module->library); /* This is a workaround for the GConf bug */

	im_ja_register_type(module);
}

void im_module_exit(void) {
	IM_JA_DEBUG("im_module_exit()\n");
}


void im_module_list(const GtkIMContextInfo ***contexts, int *n_contexts) {
	*contexts = info_list;
	*n_contexts = G_N_ELEMENTS(info_list);
}

GtkIMContext *im_module_create(const gchar *context_id) {
	IM_JA_DEBUG("im_module_create()\n");
	if (strcmp(context_id, "im-ja") == 0) return im_ja_new(); 
	else return NULL;
}

