/* -*- Mode: C; tab-width: 2; indent-tabs-mode: t; c-basic-offset: 2 -*- */
/* IM-JA Japanese Input Method Module for GTK-2.0
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Botond Botyanszki <boti@rocketmail.com>
 *          
 */

#include <config.h>

#include <gtk/gtk.h>
#include <gtk/gtkimmodule.h>

#include "im-ja.h"
#include "common.h"
#include "error.h"
#include "conf.h"
#include "nls.h"
#include "preeditwin.h"
#include "preeditarea.h"

extern IMJAConfig cfg; 
GList *preedit_windows = NULL;

static void destroy_win(PreeditWin *preedit_win, gboolean remove_signal);
//static void toplevel_destroyed_cb(GtkWidget *toplevel, GdkEvent *event, PreeditWin *preedit_win);

static void preedit_window_commit_cb(GtkWidget *button, GdkEvent *event, IMContextJA* cn) {
	IM_JA_DEBUG("preedit_window_commit_cb.\n");
	im_ja_commit(cn);
}

static void preedit_window_disable_focus_out(GtkWidget *widget, GdkEvent *event, IMContextJA* cn ) {
	cn->preedit_win->can_hide = FALSE;
}


void preedit_window_show(IMContextJA* cn) {
	IM_JA_DEBUG("preedit_window_show\n");

	if ((cn->input_method == IM_JA_RAW_INPUT) || (cn->input_method == IM_JA_KANJIPAD_INPUT)) {
		preedit_window_hide(cn);
		return;
	}

	//cn->preedit_win = g_object_get_data(G_OBJECT(cn->toplevel_gdk), "im-ja-preedit-window");
	IM_JA_DEBUG("preedit_window_create\n");
	if (cn->preedit_win == NULL) {
		GtkWidget *border;
		GtkWidget *input_border;
		GtkWidget *hbox;
		cn->preedit_win = g_new(PreeditWin, 1);
		//g_object_set_data(G_OBJECT(cn->toplevel_gdk), "im-ja-preedit-window", cn->preedit_win);
		cn->preedit_win->can_hide = TRUE;
		cn->preedit_win->window = gtk_window_new(GTK_WINDOW_POPUP);
		gtk_window_set_decorated(GTK_WINDOW(cn->preedit_win->window), FALSE);
		gtk_window_set_title(GTK_WINDOW(cn->preedit_win->window), _("[preedit window]"));
		gtk_window_set_default_size(GTK_WINDOW(cn->preedit_win->window), 140, 25);
			
		cn->preedit_win->eventbox = gtk_event_box_new();
		gtk_container_add(GTK_CONTAINER(cn->preedit_win->window), cn->preedit_win->eventbox);
		gtk_widget_show(cn->preedit_win->eventbox);
		gtk_widget_realize(cn->preedit_win->eventbox);
		gdk_window_set_cursor(cn->preedit_win->eventbox->window, gdk_cursor_new(GDK_DOT));

		/* If the preedit gets the focus, we should be still receiving keyboard input */
		g_signal_connect(G_OBJECT(cn->preedit_win->eventbox), "enter_notify_event",
										 G_CALLBACK(im_ja_grab_add), cn);
		g_signal_connect(G_OBJECT(cn->preedit_win->eventbox), "leave_notify_event",
										 G_CALLBACK(im_ja_grab_remove), cn);
		g_signal_connect(G_OBJECT(cn->preedit_win->eventbox), "button_press_event",
										 G_CALLBACK(preedit_window_commit_cb), cn);
			
		border = gtk_frame_new(NULL);
		gtk_frame_set_shadow_type(GTK_FRAME(border), GTK_SHADOW_OUT);
		gtk_container_add(GTK_CONTAINER(cn->preedit_win->eventbox), border);

		hbox = gtk_hbox_new(FALSE, 5);
		gtk_container_add(GTK_CONTAINER(border), hbox);

		input_border = gtk_frame_new(NULL);

		//cn->preedit_win->text_view = gtk_label_new(cn->preedit_buf);
		cn->preedit_win->text_view = preedit_area_new(cn->preedit_buf);

		gtk_misc_set_alignment(GTK_MISC(cn->preedit_win->text_view), 0.0, 0.5);
		gtk_container_add(GTK_CONTAINER(input_border), cn->preedit_win->text_view);
		gtk_box_pack_start(GTK_BOX(hbox), input_border, TRUE, TRUE, 0);
			
		cn->preedit_win->label = gtk_label_new(cfg.status_win_labels[cn->input_method]);
		gtk_box_pack_start(GTK_BOX(hbox), cn->preedit_win->label, FALSE, FALSE, 0);


		/* This hack is to disable flashy behaviour when window gets the focus*/
		g_signal_connect(G_OBJECT(cn->preedit_win->window), "enter_notify_event", 
										 G_CALLBACK(preedit_window_disable_focus_out), cn);
		g_signal_connect_swapped(G_OBJECT(cn->preedit_win->text_view), "realize",
														 G_CALLBACK(preedit_window_update_location), cn);
		/*
		cn->preedit_win->size_allocate_id = g_signal_connect_swapped(G_OBJECT(cn->toplevel_gtk), "size_allocate",
																																 G_CALLBACK(preedit_window_update_location), cn);
		g_signal_connect_swapped(G_OBJECT(cn->toplevel_gtk), "destroy",
														 G_CALLBACK(toplevel_destroyed_cb), cn->preedit_win);
		*/
		//cn->preedit_win->toplevel_gtk = cn->toplevel_gtk;
	}
	else {
		gtk_label_set_text(GTK_LABEL(cn->preedit_win->label), cfg.status_win_labels[cn->input_method]);
		preedit_window_set_text(cn, cn->preedit_buf); //gtk_label_set_text(GTK_LABEL(cn->preedit_win->text_view), cn->preedit_buf);
		gtk_window_resize(GTK_WINDOW(cn->preedit_win->window), 140, 25);
		preedit_window_update_location(cn);
	}
	gtk_widget_show_all(cn->preedit_win->window);
}

void preedit_window_hide(IMContextJA *cn) {
	if (cn->preedit_win != NULL) {
		IM_JA_DEBUG("preedit_window_hide\n");
		if (cn->preedit_win->can_hide == TRUE) gtk_widget_hide_all(cn->preedit_win->window);
	}
}

/*
static void toplevel_destroyed_cb(GtkWidget *toplevel, GdkEvent *event, PreeditWin *preedit_win) {
	IM_JA_DEBUG("toplevel_destroyed_cb()\n");
	destroy_win(preedit_win, FALSE);
}
*/

static void destroy_win(PreeditWin *preedit_win, gboolean remove_signal) {
	IM_JA_DEBUG("destroy_win() [%d]\n", (int) preedit_win);
	if (preedit_win == NULL) return;
	if (g_list_find(preedit_windows, preedit_win) == NULL) {
		IM_JA_DEBUG("*ERROR* preedit window is already destroyed\n");
		return;
	}
	preedit_windows = g_list_remove(preedit_windows, preedit_win);
	/*
	if (remove_signal == TRUE) {
		if (GTK_IS_WIDGET(preedit_win->toplevel_gtk)) {
			g_signal_handler_disconnect(preedit_win->toplevel_gtk, preedit_win->size_allocate_id);
		}
	}
	*/
	if (GTK_IS_WIDGET(preedit_win->window) == TRUE) gtk_widget_destroy(preedit_win->window);
	//g_free(preedit_win); this causes a segfault in some cases
}

void preedit_window_destroy(IMContextJA *cn) {
	IM_JA_DEBUG("preedit_window_destroy()\n");
	if (cn->preedit_win == NULL) return;
	destroy_win(cn->preedit_win, TRUE);
	cn->preedit_win = NULL;
}

void preedit_window_destroy_all() {
	IM_JA_DEBUG("preedit_window_destroy_all()\n");
	g_list_foreach(preedit_windows, (GFunc) destroy_win, (gpointer) TRUE);
	g_list_free(preedit_windows);
	preedit_windows = NULL;
}

void preedit_window_set_text(IMContextJA *cn, gchar *text) {
	IM_JA_DEBUG("preedit_window_set_text\n");
	preedit_area_set_text(PREEDIT_AREA(cn->preedit_win->text_view), text);	
}

void preedit_window_set_attributes(IMContextJA *cn, PangoAttrList *attrs) {
	IM_JA_DEBUG("preedit_window_set_attributes\n");
	preedit_area_set_attributes(PREEDIT_AREA(cn->preedit_win->text_view), attrs);
}

void preedit_window_set_cursor_pos(IMContextJA *cn, gint pos) {
	IM_JA_DEBUG("preedit_window_set_cursor_pos\n");
	preedit_area_set_cursor_pos(PREEDIT_AREA(cn->preedit_win->text_view), pos);
}

void preedit_window_update_location(IMContextJA *cn) {
	gint cw_x = 0, cw_y = 0; //client pos
	gint target_x = 0, target_y = 0;
	GdkRectangle toplevel_rect;

	//IM_JA_DEBUG("im_ja_update_preedit_window_location\n");
	//IM_JA_DEBUG("offsets x: %d, y: %d\n", cn->preedit_win_pos_offset_x, cn->preedit_win_pos_offset_y);
	if (cn->preedit_win !=	NULL) {
		if (GTK_IS_WIDGET(cn->toplevel_gtk) == FALSE) return;
		gdk_window_get_frame_extents(cn->toplevel_gtk->window, &toplevel_rect);

		if ((cn->preedit_win_pos_offset_x == 0) && (cn->preedit_win_pos_offset_y == 0)) {
			GtkRequisition requisition;
			gint y;
			gint screen_height = gdk_screen_get_height(gtk_widget_get_screen(cn->toplevel_gtk));
			gtk_widget_size_request(cn->preedit_win->window, &requisition);
			if (toplevel_rect.y + toplevel_rect.height + requisition.height < screen_height)  y = toplevel_rect.y + toplevel_rect.height;
			else  y = screen_height - requisition.height;
			IM_JA_DEBUG("[BOTTOM LEFT] move preedit window to x: %d, y: %d\n", toplevel_rect.x, y);
			gtk_window_move(GTK_WINDOW(cn->preedit_win->window), toplevel_rect.x, y);
		}
		else {
			if (cn->client_gdk != NULL) gdk_window_get_origin(cn->client_gdk, &cw_x, &cw_y);
			target_x = cw_x + cn->preedit_win_pos_offset_x;
			target_y = cw_y + cn->preedit_win_pos_offset_y;
			//IM_JA_DEBUG("update preedit window location x: %d, y: %d\n", target_x, target_y);
		
			if (cn->toplevel_gdk != NULL) {
				/* check if target is within the toplevel window */
				if ((toplevel_rect.width != 0) || (toplevel_rect.height != 0)) {
					if (target_y > toplevel_rect.y + toplevel_rect.height) target_y = toplevel_rect.y + toplevel_rect.height - 12;
					if (target_x > toplevel_rect.x + toplevel_rect.width) target_x = toplevel_rect.x + toplevel_rect.width;
				}
				else IM_JA_DEBUG("Couldn't get toplevel window dimensions!\n"); 
			}
			//IM_JA_DEBUG("update preedit window location x: %d, y: %d\n", target_x, target_y);
			gtk_window_move((GTK_WINDOW(cn->preedit_win->window)), target_x, target_y + 12);
		}
	}
}
