/* -*- Mode: C; tab-width: 2; indent-tabs-mode: t; c-basic-offset: 2 -*- */
/* IM-JA Japanese Input Method Module for GTK-2.0
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Botond Botyanszki <boti@rocketmail.com>
 *
 */

#include <stdio.h>
#include <string.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>


#include "im-ja.h"
#include "error.h"
#include "common.h"
#include "romakana.h"
#include "hiragana-convtable.h"
#include "zenkaku-convtable.h"
#include "halfkata-convtable.h"

extern IMJAConfig cfg;

/* * * * * * * * * * * * * * * * */

gboolean im_ja_kana_convert_tailing_n(IMContextJA *cn) {
	if (strlen(cn->preedit_buf) == 0) return FALSE;
	if (cn->preedit_buf[strlen(cn->preedit_buf) - 1] == 'n') {
		cn->preedit_buf[strlen(cn->preedit_buf) - 1] = 0;
		switch (cn->input_method) {
		case IM_JA_HIRAGANA_INPUT:
			g_strlcat(cn->preedit_buf, "\xE3\x82\x93", BUFFERSIZE);
			break;
		case IM_JA_KATAKANA_INPUT:
			g_strlcat(cn->preedit_buf, "\xE3\x83\xB3", BUFFERSIZE);
			break;
		case IM_JA_HALFKATA_INPUT:
			g_strlcat(cn->preedit_buf, "\xEF\xBE\x9D", BUFFERSIZE);
			break;
		}			
		return TRUE;
	}
	return FALSE;
}


/*
void im_ja_kana_commit_converted(IMContextJA *cn, gchar *kanastr) {
	gchar *lastkana;

	// Find any converted char 
	if (strlen(kanastr) == 0) return;
	lastkana = kanastr + strlen(kanastr);

	while (isJPChar(g_utf8_get_char(g_utf8_prev_char(lastkana))) == FALSE) {
		lastkana = g_utf8_prev_char(lastkana);
		if (lastkana == kanastr) break;
	}
	
	if (lastkana == kanastr + strlen(kanastr)) im_ja_reset(cn);
	else {   
		//copy new string back to preedit buffer 
		g_strlcpy(cn->preedit_buf, lastkana, BUFFERSIZE);
		*lastkana = 0;
	}
	if (strlen(kanastr) > 0) g_signal_emit_by_name(cn, "commit", kanastr);	
}
*/

/*
gboolean im_ja_kana_filter_keypress(IMContextJA *cn, GdkEventKey *key) {
  gchar *tmpstr, *tmp2str;

  if (key->type == GDK_KEY_RELEASE) {
    return FALSE;
  }
  if (key->keyval == GDK_space) {
    //g_strlcat(cn->preedit_buf, " ", BUFFERSIZE); 
		gtk_im_context_reset(GTK_IM_CONTEXT(cn));
		im_ja_commit_string(cn, " "); //FIXME: maybe this should be zenkaku space //ascii & unicode zenkaku diff: 65248
    return TRUE;
  }      
  
  if (key->keyval == GDK_BackSpace) {
    if (strlen(cn->preedit_buf) > 0) {
      buffer_delchar(cn->preedit_buf);
      g_signal_emit_by_name(cn, "preedit_changed");
      return TRUE;
    }
    return FALSE;
  }
  
  if ((key->keyval == GDK_Return) || (ishotkey(key, COMMIT_PREEDIT, &cfg) == TRUE)) {
    if (strlen(cn->preedit_buf) == 0) return FALSE;
		im_ja_kana_convert_tailing_n(cn);
    if (g_utf8_validate(cn->preedit_buf, -1, NULL) != TRUE) {
      printf("Cannot commit. utf8_validate failed: %s\n", cn->preedit_buf);
    }
    else g_signal_emit_by_name(cn, "commit", cn->preedit_buf);
		im_ja_reset(cn);
    return TRUE;
  } 
  
  if (key->keyval >= GDK_exclam && key->keyval <= GDK_overline) {
    gchar utf8strg[7];
		int utf8strg_end;
		utf8strg_end = g_unichar_to_utf8(gdk_keyval_to_unicode(key->keyval), utf8strg);
		utf8strg[utf8strg_end] = 0;
		g_strlcat(cn->preedit_buf, utf8strg, BUFFERSIZE);
		switch (cn->input_method) {
		case IM_JA_HIRAGANA_INPUT:
		  tmpstr = roma2kana(cn->preedit_buf, hiraganatable);
			//tmpstr = "\xE3\x82\x93";
		  IM_JA_DEBUG("converted \"%s\" to \"%s\"\n", cn->preedit_buf, tmpstr);
		  im_ja_kana_commit_converted(cn, tmpstr);
			//g_strlcpy(cn->preedit_buf, tmpstr, BUFFERSIZE);
			g_free(tmpstr);
			break;
		case IM_JA_KATAKANA_INPUT:
			tmpstr = roma2kana(cn->preedit_buf, hiraganatable);
			tmp2str = hira2kata(tmpstr);
			im_ja_kana_commit_converted(cn, tmp2str);
			g_free(tmpstr);
			g_free(tmp2str);
			break;
		case IM_JA_HALFKATA_INPUT:
			tmpstr = roma2kana(cn->preedit_buf, halfkatatable);
			im_ja_kana_commit_converted(cn, tmpstr);
			g_free(tmpstr);
			break;
		case IM_JA_ZENKAKU_INPUT:
			tmpstr = roma2kana(cn->preedit_buf, zenkakutable);
			g_strlcpy(cn->preedit_buf, tmpstr, BUFFERSIZE);
			if (g_utf8_validate(cn->preedit_buf, -1, NULL) != TRUE) {
				printf("Cannot commit. utf8_validate failed: %s\n", cn->preedit_buf);
			}
			else g_signal_emit_by_name(cn, "commit", cn->preedit_buf);
			g_free(tmpstr);
			im_ja_reset(cn);
			break;
		}
		if (g_utf8_validate(cn->preedit_buf, -1, NULL) != TRUE) {
			printf("utf8_validate failed: %s\n", cn->preedit_buf);
		}
		g_signal_emit_by_name(cn, "preedit_changed");
		return TRUE;
  }

  return FALSE; 
}
*/

int g_strrncmp(gchar* s1, size_t s1len, gchar* s2, size_t s2len) {

  if( s1len > s2len )	s1 += s1len-s2len;
  return strncmp(s1, s2, s2len);
}


/* Convert romaji to kana. Returns the kana string */
gchar *roma2kana(gchar *romastr, gint input_method) {
	gchar *result = NULL;
	gchar *tmpstr = NULL;
	romapair *romatable;
	gint i = 0, s1len, s2len;

  if (romastr == NULL) return NULL;
  if (*romastr == '\0') return NULL;

  s1len = strlen(romastr);

	switch (input_method) {
		case IM_JA_HIRAGANA_INPUT:
		case IM_JA_KATAKANA_INPUT:
			romatable = hiraganatable;
			break;
		case IM_JA_HALFKATA_INPUT:
			romatable = halfkatatable;
			break;
		case IM_JA_ZENKAKU_INPUT:
			romatable = zenkakutable;
			break;
	}

	for (i = 0; ; i++) {
    romapair rp = romatable[i];
    if (rp.kana[0] == '0' && rp.roma[0] == '0') break;
		s2len = strlen(rp.roma);
    // Compare end of input and sting in kana table
    if(g_strrncmp(romastr, s1len, rp.roma, s2len) == 0 ) {
      result = g_new0(guchar, BUFFERSIZE);
      strncat(result, romastr, s1len - s2len );
      g_strlcat(result, rp.kana, BUFFERSIZE);

			if (input_method == IM_JA_KATAKANA_INPUT) {
				tmpstr = result;
				result = hira2kata(tmpstr);
				g_free(tmpstr);
			}
			return result;
    }
  }
	return g_strdup(romastr); /* no change */
}

gchar *roma2kana_i(gchar *romastr, gint *index, gint input_method)
{
	gchar *result = NULL;
	gchar *tmpstr = NULL;
	gint i = 0,s1len,s2len;
	romapair *romatable;

	if (romastr == NULL)
		return NULL;

	result = g_new0(guchar, BUFFERSIZE);

	if (*romastr == '\0')
		return result;

	s1len = *index;

	switch (input_method) {
		case IM_JA_HIRAGANA_INPUT:
		case IM_JA_KATAKANA_INPUT:
			romatable = hiraganatable;
			break;
		case IM_JA_HALFKATA_INPUT:
			romatable = halfkatatable;
			break;
		case IM_JA_ZENKAKU_INPUT:
			romatable = zenkakutable;
			break;
	}

	for (i = 0; ; i++)
	{
		romapair rp = romatable[i];
		if (rp.kana[0] == '0' && rp.roma[0] == '0')
			break;

		s2len = strlen(rp.roma);

		// Compare offset of input and string in kana table
		if(g_strrncmp(romastr, s1len, rp.roma,s2len) == 0 )
		{
			*index += strlen(rp.kana) - s2len;
			strncat(result, romastr, s1len - s2len );
			g_strlcat(result, rp.kana, BUFFERSIZE);
			g_strlcat(result, romastr+s1len, BUFFERSIZE);
			if (input_method == IM_JA_KATAKANA_INPUT) {
				tmpstr = result;
				result = hira2kata(tmpstr);
				g_free(tmpstr);
			}
			return result;
		}
	}

	// no change
	g_strlcat(result, romastr, BUFFERSIZE);
	return result;
}


gchar *roma2kana_i_lastpass(gchar *romastr, gint *index, gint input_method)
/* IMHO the cursor position change should be done in canna_rk.c
	 depending on the return value of this function. See e.g
	 im_ja_wnn_convert_tailing_n() or im_ja_kana_convert_tailing_n()
*/
{
	gchar *result = NULL;
	gchar *subst_str;
	gchar *pattern="n";
	gint s1len;

	if (romastr == NULL)
		return NULL;

	result = g_new0(guchar, BUFFERSIZE);

	if (*romastr == '\0')
		return result;

	s1len = *index;

	switch (input_method) {
	case IM_JA_HIRAGANA_INPUT:
		subst_str = "\xE3\x82\x93";
		break;
	case IM_JA_KATAKANA_INPUT:
		subst_str = "\xE3\x83\xB3";
		break;
	case IM_JA_HALFKATA_INPUT:
		subst_str = "\xEF\xBE\x9D";
		break;
	}

	// Compare offset of input and string in kana table
	if(g_strrncmp(romastr, s1len, pattern, 1) == 0 )
	{
		*index += 3 - 1;
		strncat(result, romastr, s1len - 1 );
		g_strlcat(result, subst_str, BUFFERSIZE);
		g_strlcat(result, romastr+s1len, BUFFERSIZE);
		return result;
	}

	// no change
	g_strlcat(result, romastr, BUFFERSIZE);
	return result;
}

