/* -*- Mode: C; tab-width: 2; indent-tabs-mode: t; c-basic-offset: 2 -*- */
/* IM-JA Japanese Input Method Module for GTK-2.0
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Botond Botyanszki <boti@rocketmail.com>
 *          
 */

#include <config.h>

#include <gtk/gtk.h>
#include <gtk/gtkimmodule.h>

#include "im-ja.h"
#include "error.h"
#include "conf.h"
#include "nls.h"
#include "statuswin.h"
#include "preeditwin.h"
#include "candwin.h"

extern IMJAConfig cfg; 
GList *status_windows = NULL;
static void destroy_win(StatusWin *status_win, gboolean remove_signal);
//static void toplevel_destroyed_cb(GtkWidget *toplevel, GdkEvent *event, StatusWin *status_win);

void status_window_disable_focus_out(GtkWidget *widget, GdkEvent *event, IMContextJA* cn) {
	//IM_JA_DEBUG("status_window_disable_focus_out\n");
	cn->status_win->can_hide = FALSE;
}

void status_window_show(IMContextJA* cn) {
	IM_JA_DEBUG("status_window_show\n");

	if (cn->show_first == FALSE) return; //don't show until we get focus once
	if (candidate_window_is_shown(cn) == TRUE) return; //don't show if the candidate window is shown

	if ((cn->preedit_win_on == TRUE) && (cn->input_method != IM_JA_RAW_INPUT)) {
		// Show preedit window instead
		preedit_window_show(cn);
		return;
	}

	if (cfg.status_win_position == STATUS_DISABLED) return;
	
	if (cn->candidate_win != NULL) {
		if (GTK_WIDGET_VISIBLE(cn->candidate_win->window) == TRUE)	return;
	}
	 
	if ((cn->input_method == IM_JA_RAW_INPUT) || (cn->input_method == IM_JA_KANJIPAD_INPUT))	{
		return;
	}


	if (cn->status_win == NULL) {
		IM_JA_DEBUG("creating status win\n");
		cn->status_win = g_new(StatusWin, 1);
		status_windows = g_list_append(status_windows, cn->status_win);
		//g_object_set_data(G_OBJECT(cn->toplevel_gdk), "im-ja-status-window", cn->status_win);
		//= g_object_get_data(G_OBJECT(cn->toplevel_gdk), "im-ja-status-window");
		cn->status_win->can_hide = TRUE;
		cn->status_win->window = gtk_window_new(GTK_WINDOW_POPUP);
		gtk_window_set_title(GTK_WINDOW(cn->status_win->window), _("[status window]"));
		cn->status_win->eventbox = gtk_event_box_new();
		gtk_container_add(GTK_CONTAINER(cn->status_win->window), cn->status_win->eventbox);
		gtk_widget_show(cn->status_win->eventbox);

		/* If the preedit gets the focus, we should be still receiving keyboard input */
		g_signal_connect(G_OBJECT(cn->status_win->eventbox), "enter_notify_event",
										 G_CALLBACK(im_ja_grab_add), cn);
		g_signal_connect(G_OBJECT(cn->status_win->eventbox), "leave_notify_event",
										 G_CALLBACK(im_ja_grab_remove), cn);

		cn->status_win->border = gtk_frame_new(NULL);
		gtk_frame_set_shadow_type(GTK_FRAME(cn->status_win->border), GTK_SHADOW_OUT);
		gtk_container_add(GTK_CONTAINER(cn->status_win->eventbox), cn->status_win->border);
		cn->status_win->label = gtk_label_new(cfg.status_win_labels[cn->input_method]);
		gtk_container_add(GTK_CONTAINER(cn->status_win->border), cn->status_win->label);
		gtk_misc_set_padding(GTK_MISC(cn->status_win->label), 2, 2);

		/* This hack is to disable flashy behaviour when status window gets the focus*/
		g_signal_connect(G_OBJECT(cn->status_win->window), "enter_notify_event", 
										 G_CALLBACK(status_window_disable_focus_out), cn);
		/*			
		cn->status_win->size_allocate_id = g_signal_connect_swapped(G_OBJECT(cn->toplevel_gtk), "size_allocate",
																																G_CALLBACK(status_window_update_location), cn);
		g_signal_connect_swapped(G_OBJECT(cn->toplevel_gtk), "destroy",
														 G_CALLBACK(toplevel_destroyed_cb), cn->status_win);
		*/
		//cn->status_win->toplevel_gtk = cn->toplevel_gtk;
		g_signal_connect_swapped(G_OBJECT(cn->status_win->window), "realize",
														 G_CALLBACK(status_window_update_location), cn);
	}
	else {
		gtk_label_set_text(GTK_LABEL(cn->status_win->label), cfg.status_win_labels[cn->input_method]);
		status_window_update_location(cn);
	}
	gtk_widget_show_all(cn->status_win->window);
}

void status_window_hide(IMContextJA *cn) {
	//IM_JA_DEBUG("status_window_hide\n");
	if (cn->status_win != NULL) {
		if (cn->status_win->can_hide == TRUE) gtk_widget_hide_all(cn->status_win->window);
	}
}

/*
static void toplevel_destroyed_cb(GtkWidget *toplevel, GdkEvent *event, StatusWin *status_win) {
	IM_JA_DEBUG("toplevel_destroyed_cb()\n");
	destroy_win(status_win, FALSE);
}
*/

static void destroy_win(StatusWin *status_win, gboolean remove_signal) {
	IM_JA_DEBUG("status_win: destroy_win() [%d]\n", (int) status_win);
	if (status_win == NULL) return;
	if (g_list_find(status_windows, status_win) == NULL) {
		IM_JA_DEBUG("*ERROR* status window is already destroyed\n");
		return;
	}
	status_windows = g_list_remove(status_windows, status_win);
	/*
	if (remove_signal == TRUE) {
		if (GTK_IS_WIDGET(status_win->toplevel_gtk)) {
			g_signal_handler_disconnect(status_win->toplevel_gtk, status_win->size_allocate_id);
		}
	}
	*/
	if (GTK_IS_WIDGET(status_win->window) == TRUE) gtk_widget_destroy(status_win->window);
	//g_free(status_win); this causes a segfault in some cases
}

void status_window_destroy(IMContextJA *cn) {
	IM_JA_DEBUG("status_window_destroy()\n");
	if (cn->status_win == NULL) return;
	destroy_win(cn->status_win, TRUE);
	cn->status_win = NULL;
}

void status_window_destroy_all() {
	IM_JA_DEBUG("status_window_destroy_all()\n");
	g_list_foreach(status_windows, (GFunc) destroy_win, (gpointer) TRUE);
	g_list_free(status_windows);
	status_windows = NULL;
}


void status_window_update_location(IMContextJA *cn) {
	gint status_w, status_h;
	gint cw_x = 0, cw_y = 0; //client pos
	gint target_x = 0, target_y = 0;

	IM_JA_DEBUG("status_window_update_location [type: %d]\n", cfg.status_win_position);
	if (cn->status_win !=	NULL) {
		if (cfg.status_win_position == STATUS_DISABLED) return;
		//IM_JA_DEBUG("offsets x: %d, y: %d\n", cn->cursor_pos_offset_x, cn->cursor_pos_offset_y);

		if (cfg.status_win_position == STATUS_BELOW_CURSOR) {
			GdkRectangle toplevel_rect;
			if ((cn->cursor_pos_offset_x == 0) && (cn->cursor_pos_offset_y == 0)) goto status_win_bottom_left;
			//			if (cn->client_gdk != NULL) gdk_window_get_origin(cn->client_gdk, &cw_x, &cw_y);
			if (GDK_IS_WINDOW(cn->client_gdk)) gdk_window_get_origin(cn->client_gdk, &cw_x, &cw_y);
			target_x = cw_x + cn->cursor_pos_offset_x;
			target_y = cw_y + cn->cursor_pos_offset_y;
			if (cn->toplevel_gtk != NULL) {
				/* check if target_ is within the toplevel window */
				gdk_window_get_frame_extents(cn->toplevel_gtk->window, &toplevel_rect);
				if ((toplevel_rect.width != 0) || (toplevel_rect.height != 0)) {
					if (target_y > toplevel_rect.y + toplevel_rect.height) target_y = toplevel_rect.y + toplevel_rect.height - 12;
					if (target_x > toplevel_rect.x + toplevel_rect.width) target_x = toplevel_rect.x + toplevel_rect.width;
				}
				else IM_JA_DEBUG("Couldn't get toplevel window dimensions!\n"); 
			}
			gtk_window_get_size(GTK_WINDOW(cn->status_win->window), &status_w, &status_h);
			//IM_JA_DEBUG("[BELOW CURSOR] move status window to x: %d, y: %d\n", target_x - status_w/2, target_y + 12);
			gtk_window_move((GTK_WINDOW(cn->status_win->window)), target_x - status_w/2, target_y + 12);
			//IM_JA_DEBUG("VISIBLE: %d\n", GTK_WIDGET_VISIBLE(cn->status_win->window));
			return;
		}
		
		if (cfg.status_win_position == STATUS_WINDOW_BOTTOM_LEFT) {
		status_win_bottom_left:
			if (GTK_IS_WIDGET(cn->toplevel_gtk)) {
			//if (cn->toplevel_gtk != NULL) {
				GdkRectangle rect;
				GtkRequisition requisition;
				gint y;
				GdkScreen *screen = gtk_widget_get_screen(cn->toplevel_gtk);

				gint screen_height = gdk_screen_get_height(screen);
				gdk_window_get_frame_extents(cn->toplevel_gtk->window, &rect);
				gtk_widget_size_request(cn->status_win->window, &requisition);
				if (rect.y + rect.height + requisition.height < screen_height)  y = rect.y + rect.height;
				else  y = screen_height - requisition.height;
				//IM_JA_DEBUG("[BOTTOM LEFT] move status window to x: %d, y: %d\n", rect.x, y);
				gtk_window_move(GTK_WINDOW(cn->status_win->window), rect.x, y);
			}
			return;
		}
		
		if (cfg.status_win_position == STATUS_FIXED_COORD) {
			gtk_window_move((GTK_WINDOW(cn->status_win->window)), cfg.status_win_coords[0], cfg.status_win_coords[1]);
			return;
		}
	}
}
